@@ -24,19 +24,25 @@
 #endif
 
 #include <stdio.h>
+#include <unistd.h>
 #include <errno.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sys/ioctl.h>
+#include <sys/socket.h>
+#include <linux/if_arp.h>
+#include <linux/wireless.h>
 #include <net/ethernet.h>
 
 #include <gdbus.h>
 
 #define CONNMAN_API_SUBJECT_TO_CHANGE
 #include <connman/device.h>
+#include <connman/option.h>
+#include <connman/inet.h>
 #include <connman/dbus.h>
 #include <connman/log.h>
 
-#include "inet.h"
 #include "supplicant.h"
 
 #define TIMEOUT 5000
@@ -149,7 +155,8 @@ enum supplicant_state {
 struct supplicant_result {
 	char *path;
 	char *name;
-	char *addr;
+	unsigned char *addr;
+	unsigned int addr_len;
 	unsigned char *ssid;
 	unsigned int ssid_len;
 	dbus_uint16_t capabilities;
@@ -158,6 +165,7 @@ struct supplicant_result {
 	gboolean has_wpa;
 	gboolean has_rsn;
 	gboolean has_wps;
+	dbus_int32_t frequency;
 	dbus_int32_t quality;
 	dbus_int32_t noise;
 	dbus_int32_t level;
@@ -169,12 +177,15 @@ struct supplicant_task {
 	char *ifname;
 	struct connman_device *device;
 	struct connman_network *network;
+	struct connman_network *pending_network;
 	char *path;
 	char *netpath;
 	gboolean created;
 	enum supplicant_state state;
 	gboolean noscan;
 	GSList *scan_results;
+	struct iw_range *range;
+	gboolean disconnecting;
 };
 
 static GSList *task_list = NULL;
@@ -211,7 +222,7 @@ static struct supplicant_task *find_task_by_path(const char *path)
 	for (list = task_list; list; list = list->next) {
 		struct supplicant_task *task = list->data;
 
-		if (g_str_equal(task->path, path) == TRUE)
+		if (g_strcmp0(task->path, path) == 0)
 			return task;
 	}
 
@@ -232,7 +243,7 @@ static void add_interface_reply(DBusPendingCall *call, void *user_data)
 		return;
 
 	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
-		goto done;
+		goto failed;
 
 	dbus_error_init(&error);
 
@@ -243,7 +254,7 @@ static void add_interface_reply(DBusPendingCall *call, void *user_data)
 			dbus_error_free(&error);
 		} else
 			connman_error("Wrong arguments for add interface");
-		goto done;
+		goto failed;
 	}
 
 	DBG("path %s", path);
@@ -253,13 +264,23 @@ static void add_interface_reply(DBusPendingCall *call, void *user_data)
 
 	connman_device_set_powered(task->device, TRUE);
 
-done:
 	dbus_message_unref(reply);
+
+	return;
+
+failed:
+	task_list = g_slist_remove(task_list, task);
+
+	connman_device_unref(task->device);
+
+	free_task(task);
 }
 
 static int add_interface(struct supplicant_task *task)
 {
+	const char *driver = connman_option_get_string("wifi");
 	DBusMessage *message;
+	DBusMessageIter array, dict;
 	DBusPendingCall *call;
 
 	DBG("task %p", task);
@@ -269,8 +290,20 @@ static int add_interface(struct supplicant_task *task)
 	if (message == NULL)
 		return -ENOMEM;
 
-	dbus_message_append_args(message, DBUS_TYPE_STRING, &task->ifname,
-							DBUS_TYPE_INVALID);
+	dbus_message_iter_init_append(message, &array);
+
+	dbus_message_iter_append_basic(&array,
+					DBUS_TYPE_STRING, &task->ifname);
+
+	dbus_message_iter_open_container(&array, DBUS_TYPE_ARRAY,
+			DBUS_DICT_ENTRY_BEGIN_CHAR_AS_STRING
+			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_VARIANT_AS_STRING
+			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
+
+	connman_dbus_dict_append_variant(&dict, "driver",
+						DBUS_TYPE_STRING, &driver);
+
+	dbus_message_iter_close_container(&array, &dict);
 
 	if (dbus_connection_send_with_reply(connection, message,
 						&call, TIMEOUT) == FALSE) {
@@ -279,6 +312,12 @@ static int add_interface(struct supplicant_task *task)
 		return -EIO;
 	}
 
+	if (call == NULL) {
+		connman_error("D-Bus connection not available");
+		dbus_message_unref(message);
+		return -EIO;
+	}
+
 	dbus_pending_call_set_notify(call, add_interface_reply, task, NULL);
 
 	dbus_message_unref(message);
@@ -349,6 +388,12 @@ static int create_interface(struct supplicant_task *task)
 		return -EIO;
 	}
 
+	if (call == NULL) {
+		connman_error("D-Bus connection not available");
+		dbus_message_unref(message);
+		return -EIO;
+	}
+
 	dbus_pending_call_set_notify(call, get_interface_reply, task, NULL);
 
 	dbus_message_unref(message);
@@ -369,6 +414,8 @@ static void remove_interface_reply(DBusPendingCall *call, void *user_data)
 
 	connman_device_unref(task->device);
 
+	connman_inet_ifdown(task->ifindex);
+
 	free_task(task);
 
 	dbus_message_unref(reply);
@@ -401,6 +448,12 @@ static int remove_interface(struct supplicant_task *task)
 		return -EIO;
 	}
 
+	if (call == NULL) {
+		connman_error("D-Bus connection not available");
+		dbus_message_unref(message);
+		return -EIO;
+	}
+
 	dbus_pending_call_set_notify(call, remove_interface_reply, task, NULL);
 
 	dbus_message_unref(message);
@@ -680,23 +733,19 @@ static int set_network(struct supplicant_task *task,
 			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_VARIANT_AS_STRING
 			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
 
-	if (len > 0) {
-		if (address == NULL) {
-			const char *scan = "1";
-			connman_dbus_dict_append_variant(&dict, "scan_ssid",
-						DBUS_TYPE_STRING, &scan);
-		}
-
-		connman_dbus_dict_append_array(&dict, "ssid",
-					DBUS_TYPE_BYTE, &network, len);
-	}
-
-	if (address != NULL)
+	if (address == NULL) {
+		dbus_uint32_t scan_ssid = 1;
+		connman_dbus_dict_append_variant(&dict, "scan_ssid",
+						DBUS_TYPE_UINT32, &scan_ssid);
+	} else
 		connman_dbus_dict_append_variant(&dict, "bssid",
 						DBUS_TYPE_STRING, &address);
 
+	connman_dbus_dict_append_array(&dict, "ssid",
+					DBUS_TYPE_BYTE, &network, len);
+
 	if (g_ascii_strcasecmp(security, "wpa") == 0 ||
-				g_ascii_strcasecmp(security, "wpa2") == 0) {
+				g_ascii_strcasecmp(security, "rsn") == 0) {
 		const char *key_mgmt = "WPA-PSK";
 		connman_dbus_dict_append_variant(&dict, "key_mgmt",
 						DBUS_TYPE_STRING, &key_mgmt);
@@ -706,9 +755,13 @@ static int set_network(struct supplicant_task *task,
 						DBUS_TYPE_STRING, &passphrase);
 	} else if (g_ascii_strcasecmp(security, "wep") == 0) {
 		const char *key_mgmt = "NONE", *index = "0";
+		const char *auth_alg = "OPEN SHARED";
 		connman_dbus_dict_append_variant(&dict, "key_mgmt",
 						DBUS_TYPE_STRING, &key_mgmt);
 
+		connman_dbus_dict_append_variant(&dict, "auth_alg",
+						DBUS_TYPE_STRING, &auth_alg);
+
 		if (passphrase) {
 			int size = strlen(passphrase);
 			if (size == 10 || size == 26) {
@@ -787,6 +840,66 @@ static int initiate_scan(struct supplicant_task *task)
 	return 0;
 }
 
+static struct {
+	char *name;
+	char *value;
+} special_ssid[] = {
+	{ "<hidden>", "hidden"  },
+	{ "default",  "linksys" },
+	{ "wireless"  },
+	{ "linksys"   },
+	{ "netgear"   },
+	{ "dlink"     },
+	{ "2wire"     },
+	{ "compaq"    },
+	{ "tsunami"   },
+	{ "comcomcom", "3com"     },
+	{ "3Com",      "3com"     },
+	{ "Symbol",    "symbol"   },
+	{ "Motorola",  "motorola" },
+	{ "Wireless" , "wireless" },
+	{ "WLAN",      "wlan"     },
+	{ }
+};
+
+static char *build_group(const char *addr, const char *name,
+			const unsigned char *ssid, unsigned int ssid_len,
+					const char *mode, const char *security)
+{
+	GString *str;
+	unsigned int i;
+
+	if (addr == NULL)
+		return NULL;
+
+	str = g_string_sized_new((ssid_len * 2) + 24);
+	if (str == NULL)
+		return NULL;
+
+	for (i = 0; special_ssid[i].name; i++) {
+		if (g_strcmp0(special_ssid[i].name, name) == 0) {
+			if (special_ssid[i].value == NULL)
+				g_string_append_printf(str, "%s_%s",
+								name, addr);
+			else
+				g_string_append_printf(str, "%s_%s",
+						special_ssid[i].value, addr);
+			goto done;
+		}
+	}
+
+	if (ssid_len > 0 && ssid[0] != '\0') {
+		for (i = 0; i < ssid_len; i++)
+			g_string_append_printf(str, "%02x", ssid[i]);
+	} else
+		g_string_append_printf(str, "hidden_%s", addr);
+
+done:
+	g_string_append_printf(str, "_%s_%s", mode, security);
+
+	return g_string_free(str, FALSE);
+}
+
 static void extract_addr(DBusMessageIter *value,
 					struct supplicant_result *result)
 {
@@ -801,25 +914,20 @@ static void extract_addr(DBusMessageIter *value,
 	if (addr_len != 6)
 		return;
 
-	eth = (void *) addr;
-
-	result->addr = g_try_malloc0(18);
+	result->addr = g_try_malloc(addr_len);
 	if (result->addr == NULL)
 		return;
 
-	snprintf(result->addr, 18, "%02X:%02X:%02X:%02X:%02X:%02X",
-						eth->ether_addr_octet[0],
-						eth->ether_addr_octet[1],
-						eth->ether_addr_octet[2],
-						eth->ether_addr_octet[3],
-						eth->ether_addr_octet[4],
-						eth->ether_addr_octet[5]);
+	memcpy(result->addr, addr, addr_len);
+	result->addr_len = addr_len;
 
-	result->path = g_try_malloc0(18);
+	result->path = g_try_malloc0(13);
 	if (result->path == NULL)
 		return;
 
-	snprintf(result->path, 18, "%02X_%02X_%02X_%02X_%02X_%02X",
+	eth = (void *) addr;
+
+	snprintf(result->path, 13, "%02x%02x%02x%02x%02x%02x",
 						eth->ether_addr_octet[0],
 						eth->ether_addr_octet[1],
 						eth->ether_addr_octet[2],
@@ -833,7 +941,7 @@ static void extract_ssid(DBusMessageIter *value,
 {
 	DBusMessageIter array;
 	unsigned char *ssid;
-	int ssid_len;
+	int ssid_len, i;
 
 	dbus_message_iter_recurse(value, &array);
 	dbus_message_iter_get_fixed_array(&array, &ssid, &ssid_len);
@@ -841,9 +949,6 @@ static void extract_ssid(DBusMessageIter *value,
 	if (ssid_len < 1)
 		return;
 
-	if (ssid[0] == '\0')
-		return;
-
 	result->ssid = g_try_malloc(ssid_len);
 	if (result->ssid == NULL)
 		return;
@@ -855,7 +960,12 @@ static void extract_ssid(DBusMessageIter *value,
 	if (result->name == NULL)
 		return;
 
-	memcpy(result->name, ssid, ssid_len);
+	for (i = 0; i < ssid_len; i++) {
+		if (g_ascii_isprint(ssid[i]))
+			result->name[i] = ssid[i];
+		else
+			result->name[i] = ' ';
+	}
 }
 
 static void extract_wpaie(DBusMessageIter *value,
@@ -914,6 +1024,34 @@ static void extract_capabilites(DBusMessageIter *value,
 		result->has_wep = TRUE;
 }
 
+static unsigned char calculate_strength(struct supplicant_task *task,
+					struct supplicant_result *result)
+{
+	if (task->range->max_qual.qual == 0) {
+		unsigned char strength;
+
+		if (result->level > 0)
+			strength = 100 - result->level;
+		else
+			strength = 120 + result->level;
+
+		if (strength > 100)
+			strength = 100;
+
+		return strength;
+	}
+
+	return (result->quality * 100) / task->range->max_qual.qual;
+}
+
+static unsigned short calculate_channel(struct supplicant_result *result)
+{
+	if (result->frequency < 0)
+		return 0;
+
+	return (result->frequency - 2407) / 5;
+}
+
 static void get_properties(struct supplicant_task *task);
 
 static void properties_reply(DBusPendingCall *call, void *user_data)
@@ -923,8 +1061,10 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	struct connman_network *network;
 	DBusMessage *reply;
 	DBusMessageIter array, dict;
-	const char *mode, *security;
 	unsigned char strength;
+	unsigned short channel, frequency;
+	const char *mode, *security;
+	char *group;
 
 	DBG("task %p", task);
 
@@ -941,6 +1081,10 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	}
 
 	memset(&result, 0, sizeof(result));
+	result.frequency = -1;
+	result.quality = -1;
+	result.level = 0;
+	result.noise = 0;
 
 	dbus_message_iter_init(reply, &array);
 
@@ -986,6 +1130,8 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 			extract_wpsie(&value, &result);
 		else if (g_str_equal(key, "capabilities") == TRUE)
 			extract_capabilites(&value, &result);
+		else if (g_str_equal(key, "frequency") == TRUE)
+			dbus_message_iter_get_basic(&value, &result.frequency);
 		else if (g_str_equal(key, "quality") == TRUE)
 			dbus_message_iter_get_basic(&value, &result.quality);
 		else if (g_str_equal(key, "noise") == TRUE)
@@ -1004,10 +1150,18 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	if (result.path[0] == '\0')
 		goto done;
 
-	strength = result.quality;
+	if (result.frequency > 0 && result.frequency < 14)
+		result.frequency = 2407 + (5 * result.frequency);
+	else if (result.frequency == 14)
+		result.frequency = 2484;
+
+	strength = calculate_strength(task, &result);
+	channel  = calculate_channel(&result);
+
+	frequency = (result.frequency < 0) ? 0 : result.frequency;
 
 	if (result.has_rsn == TRUE)
-		security = "wpa2";
+		security = "rsn";
 	else if (result.has_wpa == TRUE)
 		security = "wpa";
 	else if (result.has_wep == TRUE)
@@ -1015,6 +1169,12 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	else
 		security = "none";
 
+	mode = (result.adhoc == TRUE) ? "adhoc" : "managed";
+
+	group = build_group(result.path, result.name,
+					result.ssid, result.ssid_len,
+							mode, security);
+
 	network = connman_device_get_network(task->device, result.path);
 	if (network == NULL) {
 		int index;
@@ -1030,7 +1190,8 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 		connman_network_set_protocol(network,
 						CONNMAN_NETWORK_PROTOCOL_IP);
 
-		connman_network_set_string(network, "Address", result.addr);
+		connman_network_set_address(network, result.addr,
+							result.addr_len);
 
 		if (connman_device_add_network(task->device, network) < 0) {
 			connman_network_unref(network);
@@ -1038,23 +1199,29 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 		}
 	}
 
-	connman_network_set_string(network, "Name", result.name);
+	if (result.name != NULL && result.name[0] != '\0')
+		connman_network_set_name(network, result.name);
 
 	connman_network_set_blob(network, "WiFi.SSID",
 						result.ssid, result.ssid_len);
 
-	mode = (result.adhoc == TRUE) ? "adhoc" : "managed";
 	connman_network_set_string(network, "WiFi.Mode", mode);
 
 	DBG("%s (%s %s) strength %d (%s)",
-			result.name, mode, security, strength,
-			(result.has_wps == TRUE) ? "WPS" : "no WPS");
+				result.name, mode, security, strength,
+				(result.has_wps == TRUE) ? "WPS" : "no WPS");
 
 	connman_network_set_available(network, TRUE);
-	connman_network_set_uint8(network, "Strength", strength);
+	connman_network_set_strength(network, strength);
 
+	connman_network_set_uint16(network, "Frequency", frequency);
+	connman_network_set_uint16(network, "WiFi.Channel", channel);
 	connman_network_set_string(network, "WiFi.Security", security);
 
+	connman_network_set_group(network, group);
+
+	g_free(group);
+
 done:
 	g_free(result.path);
 	g_free(result.addr);
@@ -1093,6 +1260,12 @@ static void get_properties(struct supplicant_task *task)
 		goto noscan;
 	}
 
+	if (call == NULL) {
+		connman_error("D-Bus connection not available");
+		dbus_message_unref(message);
+		goto noscan;
+	}
+
 	dbus_pending_call_set_notify(call, properties_reply, task, NULL);
 
 	dbus_message_unref(message);
@@ -1184,6 +1357,11 @@ static void scan_results_available(struct supplicant_task *task)
 	if (task->noscan == FALSE)
 		connman_device_set_scanning(task->device, TRUE);
 
+	if (call == NULL) {
+		connman_error("D-Bus connection not available");
+		goto done;
+	}
+
 	dbus_pending_call_set_notify(call, scan_results_reply, task, NULL);
 
 done:
@@ -1212,6 +1390,39 @@ static enum supplicant_state string2state(const char *state)
 		return WPA_INVALID;
 }
 
+static int task_connect(struct supplicant_task *task)
+{
+	const char *address, *security, *passphrase;
+	const void *ssid;
+	unsigned int ssid_len;
+
+	address = connman_network_get_string(task->network, "Address");
+	security = connman_network_get_string(task->network, "WiFi.Security");
+	passphrase = connman_network_get_string(task->network, "WiFi.Passphrase");
+
+	ssid = connman_network_get_blob(task->network, "WiFi.SSID", &ssid_len);
+
+	DBG("address %s security %s passphrase %s",
+					address, security, passphrase);
+
+	if (security == NULL && passphrase == NULL)
+		return -EINVAL;
+
+	if (g_str_equal(security, "none") == FALSE && passphrase == NULL)
+		return -EINVAL;
+
+	add_network(task);
+
+	select_network(task);
+	disable_network(task);
+
+	set_network(task, ssid, ssid_len, address, security, passphrase);
+
+	enable_network(task);
+
+	return 0;
+}
+
 static void state_change(struct supplicant_task *task, DBusMessage *msg)
 {
 	DBusError error;
@@ -1272,11 +1483,27 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 		connman_device_set_scanning(task->device, FALSE);
 		break;
 	case WPA_DISCONNECTED:
-		/* carrier off */
-		connman_network_set_connected(task->network, FALSE);
-		connman_device_set_scanning(task->device, FALSE);
+		if (task->disconnecting == TRUE) {
+			connman_network_set_connected(task->network, FALSE);
+			connman_network_unref(task->network);
+			task->disconnecting = FALSE;
+
+			if (task->pending_network != NULL) {
+				task->network = task->pending_network;
+				task->pending_network = NULL;
+				task_connect(task);
+			}
+		} else {
+			/* carrier off */
+			connman_network_set_connected(task->network, FALSE);
+			connman_device_set_scanning(task->device, FALSE);
+		}
+		break;
+	case WPA_ASSOCIATING:
+		connman_network_set_associating(task->network, TRUE);
 		break;
 	default:
+		connman_network_set_associating(task->network, FALSE);
 		break;
 	}
 }
@@ -1313,9 +1540,31 @@ static DBusHandlerResult supplicant_filter(DBusConnection *conn,
 	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 }
 
+static int supplicant_get_range(struct supplicant_task *task)
+{
+	struct iwreq wrq;
+	int fd, err;
+
+	fd = socket(PF_INET, SOCK_DGRAM, 0);
+	if (fd < 0)
+		return -1;
+
+	memset(&wrq, 0, sizeof(struct iwreq));
+	strncpy(wrq.ifr_name, task->ifname, IFNAMSIZ);
+	wrq.u.data.pointer = task->range;
+	wrq.u.data.length = sizeof(struct iw_range);
+
+	err = ioctl(fd, SIOCGIWRANGE, &wrq);
+
+	close(fd);
+
+	return err;
+}
+
 int supplicant_start(struct connman_device *device)
 {
 	struct supplicant_task *task;
+	int err;
 
 	DBG("device %p", device);
 
@@ -1324,22 +1573,41 @@ int supplicant_start(struct connman_device *device)
 		return -ENOMEM;
 
 	task->ifindex = connman_device_get_index(device);
-	task->ifname = inet_index2name(task->ifindex);
+	task->ifname = connman_inet_ifname(task->ifindex);
 
 	if (task->ifname == NULL) {
-		g_free(task);
-		return -ENOMEM;
+		err = -ENOMEM;
+		goto failed;
 	}
 
+	task->range = g_try_malloc0(sizeof(struct iw_range));
+	if (task->range == NULL) {
+		err = -ENOMEM;
+		goto failed;
+	}
+
+	err = supplicant_get_range(task);
+	if (err < 0)
+		goto failed;
+
 	task->device = connman_device_ref(device);
 
 	task->created = FALSE;
 	task->noscan = FALSE;
 	task->state = WPA_INVALID;
+	task->disconnecting = FALSE;
+	task->pending_network = NULL;
 
 	task_list = g_slist_append(task_list, task);
 
 	return create_interface(task);
+
+failed:
+	g_free(task->range);
+	g_free(task->ifname);
+	g_free(task);
+
+	return err;
 }
 
 int supplicant_stop(struct connman_device *device)
@@ -1353,6 +1621,8 @@ int supplicant_stop(struct connman_device *device)
 	if (task == NULL)
 		return -ENODEV;
 
+	g_free(task->range);
+
 	task_list = g_slist_remove(task_list, task);
 
 	disable_network(task);
@@ -1394,44 +1664,22 @@ int supplicant_scan(struct connman_device *device)
 int supplicant_connect(struct connman_network *network)
 {
 	struct supplicant_task *task;
-	const char *address, *security, *passphrase;
-	const void *ssid;
-	unsigned int ssid_len;
 	int index;
 
 	DBG("network %p", network);
 
-	address = connman_network_get_string(network, "Address");
-	security = connman_network_get_string(network, "WiFi.Security");
-	passphrase = connman_network_get_string(network, "WiFi.Passphrase");
-
-	ssid = connman_network_get_blob(network, "WiFi.SSID", &ssid_len);
-
-	DBG("address %s security %s passphrase %s",
-					address, security, passphrase);
-
-	if (security == NULL && passphrase == NULL)
-		return -EINVAL;
-
-	if (g_str_equal(security, "none") == FALSE && passphrase == NULL)
-		return -EINVAL;
-
 	index = connman_network_get_index(network);
 
 	task = find_task_by_index(index);
 	if (task == NULL)
 		return -ENODEV;
 
-	task->network = connman_network_ref(network);
-
-	add_network(task);
-
-	select_network(task);
-	disable_network(task);
-
-	set_network(task, ssid, ssid_len, address, security, passphrase);
-
-	enable_network(task);
+	if (task->disconnecting == TRUE)
+		task->pending_network = connman_network_ref(network);
+	else {
+		task->network = connman_network_ref(network);
+		return task_connect(task);
+	}
 
 	return 0;
 }
@@ -1449,13 +1697,14 @@ int supplicant_disconnect(struct connman_network *network)
 	if (task == NULL)
 		return -ENODEV;
 
+	if (task->disconnecting == TRUE)
+		return -EINPROGRESS;
+
 	disable_network(task);
 
 	remove_network(task);
 
-	connman_network_set_connected(task->network, FALSE);
-
-	connman_network_unref(task->network);
+	task->disconnecting = TRUE;
 
 	return 0;
 }
