@@ -40,14 +40,20 @@ struct connman_service {
 	enum connman_service_mode mode;
 	enum connman_service_security security;
 	enum connman_service_state state;
+	enum connman_service_error error;
 	connman_uint8_t strength;
 	connman_bool_t favorite;
+	connman_bool_t hidden;
+	GTimeVal modified;
 	unsigned int order;
 	char *name;
 	char *passphrase;
 	char *profile;
+	struct connman_ipconfig *ipconfig;
 	struct connman_device *device;
 	struct connman_network *network;
+	DBusMessage *pending;
+	guint timeout;
 };
 
 static void append_path(gpointer value, gpointer user_data)
@@ -69,6 +75,34 @@ void __connman_service_list(DBusMessageIter *iter)
 	g_sequence_foreach(service_list, append_path, iter);
 }
 
+struct find_data {
+	const char *path;
+	struct connman_service *service;
+};
+
+static void compare_path(gpointer value, gpointer user_data)
+{
+	struct connman_service *service = value;
+	struct find_data *data = user_data;
+
+	if (data->service != NULL)
+		return;
+
+	if (g_strcmp0(service->path, data->path) == 0)
+		data->service = service;
+}
+
+static struct connman_service *find_service(const char *path)
+{
+	struct find_data data = { .path = path, .service = NULL };
+
+	DBG("path %s", path);
+
+	g_sequence_foreach(service_list, compare_path, &data);
+
+	return data.service;
+}
+
 static const char *type2string(enum connman_service_type type)
 {
 	switch (type) {
@@ -80,6 +114,10 @@ static const char *type2string(enum connman_service_type type)
 		return "wifi";
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 		return "wimax";
+	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+		return "bluetooth";
+	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		return "cellular";
 	}
 
 	return NULL;
@@ -110,8 +148,8 @@ static const char *security2string(enum connman_service_security security)
 		return "wep";
 	case CONNMAN_SERVICE_SECURITY_WPA:
 		return "wpa";
-	case CONNMAN_SERVICE_SECURITY_WPA2:
-		return "wpa2";
+	case CONNMAN_SERVICE_SECURITY_RSN:
+		return "rsn";
 	}
 
 	return NULL;
@@ -141,6 +179,30 @@ static const char *state2string(enum connman_service_state state)
 	return NULL;
 }
 
+static const char *error2string(enum connman_service_error error)
+{
+	switch (error) {
+	case CONNMAN_SERVICE_ERROR_UNKNOWN:
+		break;
+	case CONNMAN_SERVICE_ERROR_DHCP_FAILED:
+		return "dhcp-failed";
+	case CONNMAN_SERVICE_ERROR_PIN_MISSING:
+		return "pin-missing";
+	}
+
+	return NULL;
+}
+
+static enum connman_service_error string2error(const char *error)
+{
+	if (g_strcmp0(error, "dhcp-failed") == 0)
+		return CONNMAN_SERVICE_ERROR_DHCP_FAILED;
+	else if (g_strcmp0(error, "pin-missing") == 0)
+		return CONNMAN_SERVICE_ERROR_PIN_MISSING;
+
+	return CONNMAN_SERVICE_ERROR_UNKNOWN;
+}
+
 static void state_changed(struct connman_service *service)
 {
 	DBusMessage *signal;
@@ -171,15 +233,45 @@ static void state_changed(struct connman_service *service)
 	g_dbus_send_message(connection, signal);
 }
 
-static DBusMessage *get_properties(DBusConnection *conn,
-					DBusMessage *msg, void *data)
+static void strength_changed(struct connman_service *service)
 {
-	struct connman_service *service = data;
+	DBusMessage *signal;
+	DBusMessageIter entry, value;
+	const char *key = "Strength";
+
+	if (service->path == NULL)
+		return;
+
+	if (service->strength == 0)
+		return;
+
+	signal = dbus_message_new_signal(service->path,
+				CONNMAN_SERVICE_INTERFACE, "PropertyChanged");
+	if (signal == NULL)
+		return;
+
+	dbus_message_iter_init_append(signal, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+					DBUS_TYPE_BYTE_AS_STRING, &value);
+	dbus_message_iter_append_basic(&value, DBUS_TYPE_BYTE,
+							&service->strength);
+	dbus_message_iter_close_container(&entry, &value);
+
+	g_dbus_send_message(connection, signal);
+}
+
+static DBusMessage *get_properties(DBusConnection *conn,
+					DBusMessage *msg, void *user_data)
+{
+	struct connman_service *service = user_data;
 	DBusMessage *reply;
 	DBusMessageIter array, dict;
 	const char *str;
 
-	DBG("conn %p", conn);
+	DBG("service %p", service);
 
 	reply = dbus_message_new_method_return(msg);
 	if (reply == NULL)
@@ -212,6 +304,11 @@ static DBusMessage *get_properties(DBusConnection *conn,
 		connman_dbus_dict_append_variant(&dict, "State",
 						DBUS_TYPE_STRING, &str);
 
+	str = error2string(service->error);
+	if (str != NULL)
+		connman_dbus_dict_append_variant(&dict, "Error",
+						DBUS_TYPE_STRING, &str);
+
 	if (service->strength > 0)
 		connman_dbus_dict_append_variant(&dict, "Strength",
 					DBUS_TYPE_BYTE, &service->strength);
@@ -229,20 +326,22 @@ static DBusMessage *get_properties(DBusConnection *conn,
 		connman_dbus_dict_append_variant(&dict, "Passphrase",
 				DBUS_TYPE_STRING, &service->passphrase);
 
+	__connman_ipconfig_append_ipv4(service->ipconfig, &dict, "IPv4.");
+
 	dbus_message_iter_close_container(&array, &dict);
 
 	return reply;
 }
 
 static DBusMessage *set_property(DBusConnection *conn,
-					DBusMessage *msg, void *data)
+					DBusMessage *msg, void *user_data)
 {
-	struct connman_service *service = data;
+	struct connman_service *service = user_data;
 	DBusMessageIter iter, value;
 	const char *name;
 	int type;
 
-	DBG("conn %p", conn);
+	DBG("service %p", service);
 
 	if (dbus_message_iter_init(msg, &iter) == FALSE)
 		return __connman_error_invalid_arguments(msg);
@@ -277,118 +376,330 @@ static DBusMessage *set_property(DBusConnection *conn,
 				"WiFi.Passphrase", service->passphrase);
 
 		__connman_storage_save_service(service);
+	} else if (g_str_has_prefix(name, "IPv4.") == TRUE) {
+		int err;
+
+		err = __connman_ipconfig_set_ipv4(service->ipconfig,
+							name + 5, &value);
+		if (err < 0)
+			return __connman_error_failed(msg, -err);
+	} else
+		return __connman_error_invalid_property(msg);
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+}
+
+static DBusMessage *clear_property(DBusConnection *conn,
+					DBusMessage *msg, void *user_data)
+{
+	struct connman_service *service = user_data;
+	const char *name;
+
+	DBG("service %p", service);
+
+	dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &name,
+							DBUS_TYPE_INVALID);
+
+	if (__connman_security_check_privilege(msg,
+					CONNMAN_SECURITY_PRIVILEGE_MODIFY) < 0)
+		return __connman_error_permission_denied(msg);
+
+	if (g_str_equal(name, "Error") == TRUE) {
+		service->state = CONNMAN_SERVICE_STATE_IDLE;
+		service->error = CONNMAN_SERVICE_ERROR_UNKNOWN;
+		state_changed(service);
+
+		g_get_current_time(&service->modified);
+		__connman_storage_save_service(service);
+	} else
+		return __connman_error_invalid_property(msg);
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+}
+
+static connman_bool_t is_connecting(struct connman_service *service)
+{
+	switch (service->state) {
+	case CONNMAN_SERVICE_STATE_UNKNOWN:
+	case CONNMAN_SERVICE_STATE_IDLE:
+	case CONNMAN_SERVICE_STATE_CARRIER:
+	case CONNMAN_SERVICE_STATE_FAILURE:
+	case CONNMAN_SERVICE_STATE_DISCONNECT:
+	case CONNMAN_SERVICE_STATE_READY:
+		break;
+	case CONNMAN_SERVICE_STATE_ASSOCIATION:
+	case CONNMAN_SERVICE_STATE_CONFIGURATION:
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
+static struct connman_service *find_pending_service(void)
+{
+	struct connman_service *service;
+	GSequenceIter *iter;
+
+	iter = g_sequence_get_begin_iter(service_list);
+
+	while (g_sequence_iter_is_end(iter) == FALSE) {
+		service = g_sequence_get(iter);
+		if (service->pending != NULL)
+			return service;
+
+		iter = g_sequence_iter_next(iter);
+	}
+
+	return NULL;
+}
+
+static void __connman_service_auto_connect(void)
+{
+	struct connman_service *service;
+	GSequenceIter *iter;
+
+	DBG("");
+
+	service = find_pending_service();
+	if (service != NULL)
+		return;
+
+	iter = g_sequence_get_begin_iter(service_list);
+	if (g_sequence_iter_is_end(iter) == TRUE)
+		return;
+
+	service = g_sequence_get(iter);
+
+	while (service->state == CONNMAN_SERVICE_STATE_FAILURE) {
+		iter = g_sequence_iter_next(iter);
+		if (g_sequence_iter_is_end(iter))
+			return;
+		service = g_sequence_get(iter);
+	}
+
+	if (service->favorite == FALSE)
+		return;
+
+	if (service->state == CONNMAN_SERVICE_STATE_READY)
+		return;
+
+	if (is_connecting(service) == TRUE)
+		return;
+
+	if (service->state == CONNMAN_SERVICE_STATE_IDLE)
+		__connman_service_connect(service);
+}
+
+static gboolean connect_timeout(gpointer user_data)
+{
+	struct connman_service *service = user_data;
+	connman_bool_t auto_connect = FALSE;
+
+	DBG("service %p", service);
+
+	service->timeout = 0;
+
+	if (service->network != NULL) {
+		connman_bool_t connected;
+
+		connected = connman_network_get_connected(service->network);
+		if (connected == TRUE) {
+			__connman_service_indicate_state(service,
+						CONNMAN_SERVICE_STATE_READY);
+			return FALSE;
+		}
+
+		__connman_network_disconnect(service->network);
+	} else if (service->device != NULL) {
+		connman_bool_t disconnected;
+
+		disconnected = connman_device_get_disconnected(service->device);
+		if (disconnected == FALSE) {
+			__connman_service_indicate_state(service,
+						CONNMAN_SERVICE_STATE_READY);
+			return FALSE;
+		}
+
+		__connman_device_disconnect(service->device);
+	}
+
+	if (service->pending != NULL) {
+		DBusMessage *reply;
+
+		reply = __connman_error_operation_timeout(service->pending);
+		if (reply != NULL)
+			g_dbus_send_message(connection, reply);
+
+		dbus_message_unref(service->pending);
+		service->pending = NULL;
+	} else
+		auto_connect = TRUE;
+
+	__connman_service_indicate_state(service,
+					CONNMAN_SERVICE_STATE_FAILURE);
+
+	if (auto_connect == TRUE)
+		__connman_service_auto_connect();
+
+	return FALSE;
+}
+
+static DBusMessage *connect_service(DBusConnection *conn,
+					DBusMessage *msg, void *user_data)
+{
+	struct connman_service *service = user_data;
+	int err;
+
+	DBG("service %p", service);
+
+	if (service->pending != NULL)
+		return __connman_error_in_progress(msg);
+
+	service->pending = dbus_message_ref(msg);
+
+	err = __connman_service_connect(service);
+	if (err < 0) {
+		if (err != -EINPROGRESS) {
+			dbus_message_unref(service->pending);
+			service->pending = NULL;
+
+			return __connman_error_failed(msg, -err);
+		}
+
+		return NULL;
+	}
+
+	dbus_message_unref(service->pending);
+	service->pending = NULL;
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+}
+
+static DBusMessage *disconnect_service(DBusConnection *conn,
+					DBusMessage *msg, void *user_data)
+{
+	struct connman_service *service = user_data;
+	int err;
+
+	DBG("service %p", service);
+
+	if (service->pending != NULL) {
+		DBusMessage *reply;
+
+		reply = __connman_error_operation_aborted(service->pending);
+		if (reply != NULL)
+			g_dbus_send_message(conn, reply);
+
+		dbus_message_unref(service->pending);
+		service->pending = NULL;
+
+		return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+	}
+
+	err = __connman_service_disconnect(service);
+	if (err < 0) {
+		if (err != -EINPROGRESS)
+			return __connman_error_failed(msg, -err);
+
+		return NULL;
 	}
 
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
-static DBusMessage *connect_service(DBusConnection *conn,
-					DBusMessage *msg, void *data)
-{
-	struct connman_service *service = data;
-
-	if (service->network != NULL) {
-		int err;
-
-		connman_network_set_string(service->network,
-				"WiFi.Passphrase", service->passphrase);
-
-		err = connman_network_connect(service->network);
-		if (err < 0 && err != -EINPROGRESS)
-			return __connman_error_failed(msg, -err);
-
-		return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
-	} else if (service->device != NULL) {
-		if (service->favorite == FALSE)
-			return __connman_error_no_carrier(msg);
-
-		if (__connman_device_connect(service->device) < 0)
-			return __connman_error_failed(msg, EINVAL);
-
-		return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
-	}
-
-	return __connman_error_not_supported(msg);
-}
-
-static DBusMessage *disconnect_service(DBusConnection *conn,
-					DBusMessage *msg, void *data)
-{
-	struct connman_service *service = data;
-
-	if (service->network != NULL) {
-		int err;
-
-		err = __connman_network_disconnect(service->network);
-		if (err < 0 && err != -EINPROGRESS)
-			return __connman_error_failed(msg, -err);
-
-		return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
-	} else if (service->device != NULL) {
-		int err;
-
-		if (service->favorite == FALSE)
-			return __connman_error_no_carrier(msg);
-
-		err = __connman_device_disconnect(service->device);
-		if (err < 0)
-			return __connman_error_failed(msg, -err);
-
-		return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
-	}
-
-	return __connman_error_not_supported(msg);
-}
-
 static DBusMessage *remove_service(DBusConnection *conn,
-					DBusMessage *msg, void *data)
+					DBusMessage *msg, void *user_data)
 {
-	struct connman_service *service = data;
+	struct connman_service *service = user_data;
+
+	DBG("service %p", service);
 
 	if (service->type == CONNMAN_SERVICE_TYPE_ETHERNET)
 		return __connman_error_not_supported(msg);
 
-	if (service->network != NULL) {
-		int err;
+	if (service->favorite == FALSE)
+		return __connman_error_not_supported(msg);
 
-		err = __connman_network_disconnect(service->network);
-		if (err < 0 && err != -EINPROGRESS)
-			return __connman_error_failed(msg, -err);
-	}
+	if (service->network != NULL)
+		__connman_network_disconnect(service->network);
+
+	g_free(service->passphrase);
+	service->passphrase = NULL;
 
 	connman_service_set_favorite(service, FALSE);
-
 	__connman_storage_save_service(service);
 
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
 static DBusMessage *move_before(DBusConnection *conn,
-					DBusMessage *msg, void *data)
+					DBusMessage *msg, void *user_data)
 {
-	struct connman_service *service = data;
+	struct connman_service *service = user_data;
+	struct connman_service *target;
+	const char *path;
+	GSequenceIter *src, *dst;
+
+	DBG("service %p", service);
+
+	dbus_message_get_args(msg, NULL, DBUS_TYPE_OBJECT_PATH, &path,
+							DBUS_TYPE_INVALID);
 
 	if (service->favorite == FALSE)
 		return __connman_error_not_supported(msg);
 
-	return __connman_error_not_implemented(msg);
+	target = find_service(path);
+	if (target == NULL || target->favorite == FALSE || target == service)
+		return __connman_error_invalid_service(msg);
+
+	DBG("target %s", target->identifier);
+
+	g_get_current_time(&service->modified);
+	__connman_storage_save_service(service);
+
+	src = g_hash_table_lookup(service_hash, service->identifier);
+	dst = g_hash_table_lookup(service_hash, target->identifier);
+
+	g_sequence_move(src, dst);
+
+	__connman_profile_changed();
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
 static DBusMessage *move_after(DBusConnection *conn,
-					DBusMessage *msg, void *data)
+					DBusMessage *msg, void *user_data)
 {
-	struct connman_service *service = data;
+	struct connman_service *service = user_data;
+	struct connman_service *target;
+	const char *path;
+
+	DBG("service %p", service);
+
+	dbus_message_get_args(msg, NULL, DBUS_TYPE_OBJECT_PATH, &path,
+							DBUS_TYPE_INVALID);
 
 	if (service->favorite == FALSE)
 		return __connman_error_not_supported(msg);
 
+	target = find_service(path);
+	if (target == NULL || target->favorite == FALSE || target == service)
+		return __connman_error_invalid_service(msg);
+
+	DBG("target %s", target->identifier);
+
+	g_get_current_time(&service->modified);
+	__connman_storage_save_service(service);
+
 	return __connman_error_not_implemented(msg);
 }
 
 static GDBusMethodTable service_methods[] = {
 	{ "GetProperties", "",   "a{sv}", get_properties     },
 	{ "SetProperty",   "sv", "",      set_property       },
-	{ "Connect",       "",   "",      connect_service    },
+	{ "ClearProperty", "s",  "",      clear_property     },
+	{ "Connect",       "",   "",      connect_service,
+						G_DBUS_METHOD_FLAG_ASYNC },
 	{ "Disconnect",    "",   "",      disconnect_service },
 	{ "Remove",        "",   "",      remove_service     },
 	{ "MoveBefore",    "o",  "",      move_before        },
@@ -401,15 +712,25 @@ static GDBusSignalTable service_signals[] = {
 	{ },
 };
 
-static void service_free(gpointer data)
+static void service_free(gpointer user_data)
 {
-	struct connman_service *service = data;
+	struct connman_service *service = user_data;
 	char *path = service->path;
 
 	DBG("service %p", service);
 
 	g_hash_table_remove(service_hash, service->identifier);
 
+	if (service->timeout > 0) {
+		g_source_remove(service->timeout);
+		service->timeout = 0;
+	}
+
+	if (service->pending != NULL) {
+		dbus_message_unref(service->pending);
+		service->pending = NULL;
+	}
+
 	service->path = NULL;
 
 	if (path != NULL) {
@@ -423,6 +744,8 @@ static void service_free(gpointer data)
 	if (service->network != NULL)
 		connman_network_unref(service->network);
 
+	connman_ipconfig_unref(service->ipconfig);
+
 	g_free(service->profile);
 	g_free(service->name);
 	g_free(service->passphrase);
@@ -431,12 +754,12 @@ static void service_free(gpointer data)
 }
 
 /**
- * connman_service_put:
+ * __connman_service_put:
  * @service: service structure
  *
  * Release service if no longer needed
  */
-void connman_service_put(struct connman_service *service)
+void __connman_service_put(struct connman_service *service)
 {
 	DBG("service %p", service);
 
@@ -463,6 +786,7 @@ static void __connman_service_initialize(struct connman_service *service)
 	service->state    = CONNMAN_SERVICE_STATE_UNKNOWN;
 
 	service->favorite = FALSE;
+	service->hidden = FALSE;
 
 	service->order = 0;
 }
@@ -486,6 +810,15 @@ struct connman_service *connman_service_create(void)
 
 	__connman_service_initialize(service);
 
+	service->ipconfig = connman_ipconfig_create();
+	if (service->ipconfig == NULL) {
+		g_free(service);
+		return NULL;
+	}
+
+	connman_ipconfig_set_method(service->ipconfig,
+					CONNMAN_IPCONFIG_METHOD_DHCP);
+
 	return service;
 }
 
@@ -510,7 +843,7 @@ struct connman_service *connman_service_ref(struct connman_service *service)
  */
 void connman_service_unref(struct connman_service *service)
 {
-	connman_service_put(service);
+	__connman_service_put(service);
 }
 
 static gint service_compare(gconstpointer a, gconstpointer b,
@@ -519,6 +852,13 @@ static gint service_compare(gconstpointer a, gconstpointer b,
 	struct connman_service *service_a = (void *) a;
 	struct connman_service *service_b = (void *) b;
 
+	if (service_a->state != service_b->state) {
+		if (service_a->state == CONNMAN_SERVICE_STATE_READY)
+			return -1;
+		if (service_b->state == CONNMAN_SERVICE_STATE_READY)
+			return 1;
+	}
+
 	if (service_a->order > service_b->order)
 		return -1;
 
@@ -550,7 +890,7 @@ int connman_service_set_favorite(struct connman_service *service,
 	if (iter == NULL)
 		return -ENOENT;
 
-	if (service->favorite)
+	if (service->favorite == favorite)
 		return -EALREADY;
 
 	service->favorite = favorite;
@@ -574,6 +914,8 @@ int __connman_service_set_carrier(struct connman_service *service,
 	case CONNMAN_SERVICE_TYPE_UNKNOWN:
 	case CONNMAN_SERVICE_TYPE_WIFI:
 	case CONNMAN_SERVICE_TYPE_WIMAX:
+	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+	case CONNMAN_SERVICE_TYPE_CELLULAR:
 		return -EINVAL;
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
 		break;
@@ -596,6 +938,8 @@ int __connman_service_set_carrier(struct connman_service *service,
 int __connman_service_indicate_state(struct connman_service *service,
 					enum connman_service_state state)
 {
+	GSequenceIter *iter;
+
 	DBG("service %p state %d", service, state);
 
 	if (service == NULL)
@@ -620,19 +964,148 @@ int __connman_service_indicate_state(struct connman_service *service,
 	service->state = state;
 	state_changed(service);
 
-	if (state == CONNMAN_SERVICE_STATE_READY)
-		return connman_service_set_favorite(service, TRUE);
+	if (state == CONNMAN_SERVICE_STATE_READY) {
+		connman_service_set_favorite(service, TRUE);
+
+		if (service->timeout > 0) {
+			g_source_remove(service->timeout);
+			service->timeout = 0;
+		}
+
+		if (service->pending != NULL) {
+			g_dbus_send_reply(connection, service->pending,
+							DBUS_TYPE_INVALID);
+
+			dbus_message_unref(service->pending);
+			service->pending = NULL;
+		}
+
+		g_get_current_time(&service->modified);
+		__connman_storage_save_service(service);
+	}
+
+	if (state == CONNMAN_SERVICE_STATE_FAILURE) {
+		if (service->timeout > 0) {
+			g_source_remove(service->timeout);
+			service->timeout = 0;
+		}
+
+		if (service->pending != NULL) {
+			DBusMessage *reply;
+
+			reply = __connman_error_failed(service->pending, EIO);
+			if (reply != NULL)
+				g_dbus_send_message(connection, reply);
+
+			dbus_message_unref(service->pending);
+			service->pending = NULL;
+		}
+
+		g_get_current_time(&service->modified);
+		__connman_storage_save_service(service);
+	} else
+		service->error = CONNMAN_SERVICE_ERROR_UNKNOWN;
+
+	iter = g_hash_table_lookup(service_hash, service->identifier);
+	if (iter != NULL)
+		g_sequence_sort_changed(iter, service_compare, NULL);
+
+	__connman_profile_changed();
+
+	return 0;
+}
+
+int __connman_service_indicate_error(struct connman_service *service,
+					enum connman_service_error error)
+{
+	DBG("service %p error %d", service, error);
+
+	if (service == NULL)
+		return -EINVAL;
+
+	service->error = error;
+
+	return __connman_service_indicate_state(service,
+					CONNMAN_SERVICE_STATE_FAILURE);
+}
+
+int __connman_service_indicate_default(struct connman_service *service)
+{
+	DBG("service %p", service);
+
+	return 0;
+}
+
+int __connman_service_connect(struct connman_service *service)
+{
+	int err;
+
+	DBG("service %p", service);
+
+	if (service->state == CONNMAN_SERVICE_STATE_READY)
+		return -EISCONN;
+
+	if (service->network != NULL) {
+		if (service->hidden == TRUE)
+			return -EINVAL;
+
+		connman_network_set_string(service->network,
+				"WiFi.Passphrase", service->passphrase);
+
+		err = __connman_network_connect(service->network);
+	} else if (service->device != NULL) {
+		if (service->favorite == FALSE)
+			return -ENOLINK;
+
+		err = __connman_device_connect(service->device);
+	} else
+		return -EOPNOTSUPP;
+
+	if (err < 0) {
+		if (err != -EINPROGRESS)
+			return err;
+
+		service->timeout = g_timeout_add_seconds(45,
+						connect_timeout, service);
+
+		return -EINPROGRESS;
+	}
+
+	return 0;
+}
+
+int __connman_service_disconnect(struct connman_service *service)
+{
+	int err;
+
+	DBG("service %p", service);
+
+	if (service->network != NULL) {
+		err = __connman_network_disconnect(service->network);
+	} else if (service->device != NULL) {
+		if (service->favorite == FALSE)
+			return -ENOLINK;
+		err = __connman_device_disconnect(service->device);
+	} else
+		return -EOPNOTSUPP;
+
+	if (err < 0) {
+		if (err != -EINPROGRESS)
+			return err;
+
+		return -EINPROGRESS;
+	}
 
 	return 0;
 }
 
 /**
- * connman_service_lookup:
+ * __connman_service_lookup:
  * @identifier: service identifier
  *
  * Look up a service by identifier (reference count will not be increased)
  */
-struct connman_service *connman_service_lookup(const char *identifier)
+static struct connman_service *__connman_service_lookup(const char *identifier)
 {
 	GSequenceIter *iter;
 
@@ -644,12 +1117,12 @@ struct connman_service *connman_service_lookup(const char *identifier)
 }
 
 /**
- * connman_service_get:
+ * __connman_service_get:
  * @identifier: service identifier
  *
  * Look up a service by identifier or create a new one if not found
  */
-struct connman_service *connman_service_get(const char *identifier)
+static struct connman_service *__connman_service_get(const char *identifier)
 {
 	struct connman_service *service;
 	GSequenceIter *iter;
@@ -662,14 +1135,12 @@ struct connman_service *connman_service_get(const char *identifier)
 		return service;
 	}
 
-	service = g_try_new0(struct connman_service, 1);
+	service = connman_service_create();
 	if (service == NULL)
 		return NULL;
 
 	DBG("service %p", service);
 
-	__connman_service_initialize(service);
-
 	service->identifier = g_strdup(identifier);
 
 	service->profile = g_strdup(__connman_profile_active_ident());
@@ -687,6 +1158,7 @@ struct connman_service *connman_service_get(const char *identifier)
 static int service_register(struct connman_service *service)
 {
 	const char *path = __connman_profile_active_path();
+	GSequenceIter *iter;
 
 	DBG("service %p", service);
 
@@ -702,6 +1174,12 @@ static int service_register(struct connman_service *service)
 					service_methods, service_signals,
 							NULL, service, NULL);
 
+	__connman_storage_load_service(service);
+
+	iter = g_hash_table_lookup(service_hash, service->identifier);
+	if (iter != NULL)
+		g_sequence_sort_changed(iter, service_compare, NULL);
+
 	__connman_profile_changed();
 
 	return 0;
@@ -726,7 +1204,7 @@ struct connman_service *__connman_service_lookup_from_device(struct connman_devi
 	name = g_strdup_printf("%s_%s",
 				__connman_device_get_type(device), ident);
 
-	service = connman_service_lookup(name);
+	service = __connman_service_lookup(name);
 
 	g_free(name);
 
@@ -744,6 +1222,7 @@ static enum connman_service_type convert_device_type(struct connman_device *devi
 	case CONNMAN_DEVICE_TYPE_WIMAX:
 	case CONNMAN_DEVICE_TYPE_BLUETOOTH:
 	case CONNMAN_DEVICE_TYPE_GPS:
+	case CONNMAN_DEVICE_TYPE_MBM:
 	case CONNMAN_DEVICE_TYPE_HSO:
 	case CONNMAN_DEVICE_TYPE_NOZOMI:
 	case CONNMAN_DEVICE_TYPE_HUAWEI:
@@ -775,12 +1254,12 @@ struct connman_service *__connman_service_create_from_device(struct connman_devi
 	name = g_strdup_printf("%s_%s",
 				__connman_device_get_type(device), ident);
 
-	service = connman_service_get(name);
+	service = __connman_service_get(name);
 	if (service == NULL)
 		goto done;
 
 	if (service->path != NULL) {
-		connman_service_put(service);
+		__connman_service_put(service);
 		service = NULL;
 		goto done;
 	}
@@ -791,6 +1270,9 @@ struct connman_service *__connman_service_create_from_device(struct connman_devi
 
 	service_register(service);
 
+	if (service->favorite == TRUE)
+		__connman_service_auto_connect();
+
 done:
 	g_free(name);
 
@@ -820,13 +1302,21 @@ struct connman_service *__connman_service_lookup_from_network(struct connman_net
 	name = g_strdup_printf("%s_%s_%s",
 			__connman_network_get_type(network), ident, group);
 
-	service = connman_service_lookup(name);
+	service = __connman_service_lookup(name);
 
 	g_free(name);
 
 	return service;
 }
 
+unsigned int __connman_service_get_order(struct connman_service *service)
+{
+	if (service == NULL)
+		return 0;
+
+	return service->order;
+}
+
 static enum connman_service_type convert_network_type(struct connman_network *network)
 {
 	enum connman_network_type type = connman_network_get_type(network);
@@ -869,8 +1359,8 @@ static enum connman_service_mode convert_wifi_security(const char *security)
 		return CONNMAN_SERVICE_SECURITY_WEP;
 	else if (g_str_equal(security, "wpa") == TRUE)
 		return CONNMAN_SERVICE_SECURITY_WPA;
-	else if (g_str_equal(security, "wpa2") == TRUE)
-		return CONNMAN_SERVICE_SECURITY_WPA2;
+	else if (g_str_equal(security, "rsn") == TRUE)
+		return CONNMAN_SERVICE_SECURITY_RSN;
 	else
 		return CONNMAN_SERVICE_SECURITY_UNKNOWN;
 }
@@ -882,10 +1372,21 @@ static void update_from_network(struct connman_service *service,
 	GSequenceIter *iter;
 	const char *str;
 
+	if (service->state == CONNMAN_SERVICE_STATE_READY)
+		return;
+
+	if (is_connecting(service) == TRUE)
+		return;
+
 	str = connman_network_get_string(network, "Name");
 	if (str != NULL) {
 		g_free(service->name);
 		service->name = g_strdup(str);
+		service->hidden = FALSE;
+	} else {
+		g_free(service->name);
+		service->name = NULL;
+		service->hidden = TRUE;
 	}
 
 	service->strength = connman_network_get_uint8(network, "Strength");
@@ -899,6 +1400,8 @@ static void update_from_network(struct connman_service *service,
 	if (service->strength > strength && service->network != NULL) {
 		connman_network_unref(service->network);
 		service->network = NULL;
+
+		strength_changed(service);
 	}
 
 	if (service->network == NULL) {
@@ -928,6 +1431,12 @@ struct connman_service *__connman_service_create_from_network(struct connman_net
 	const char *ident, *group;
 	char *name;
 
+	service = __connman_service_lookup_from_network(network);
+	if (service != NULL) {
+		update_from_network(service, network);
+		return service;
+	}
+
 	ident = __connman_network_get_ident(network);
 	if (ident == NULL)
 		return NULL;
@@ -939,7 +1448,7 @@ struct connman_service *__connman_service_create_from_network(struct connman_net
 	name = g_strdup_printf("%s_%s_%s",
 			__connman_network_get_type(network), ident, group);
 
-	service = connman_service_get(name);
+	service = __connman_service_get(name);
 	if (service == NULL)
 		goto done;
 
@@ -948,17 +1457,22 @@ struct connman_service *__connman_service_create_from_network(struct connman_net
 
 		__connman_profile_changed();
 
-		connman_service_put(service);
+		__connman_service_put(service);
 		service = NULL;
 		goto done;
 	}
 
 	service->type = convert_network_type(network);
 
+	service->state = CONNMAN_SERVICE_STATE_IDLE;
+
 	update_from_network(service, network);
 
 	service_register(service);
 
+	if (service->favorite == TRUE)
+		__connman_service_auto_connect();
+
 done:
 	g_free(name);
 
@@ -970,7 +1484,7 @@ static int service_load(struct connman_service *service)
 	GKeyFile *keyfile;
 	gchar *pathname, *data = NULL;
 	gsize length;
-	char *str;
+	gchar *str;
 
 	DBG("service %p", service);
 
@@ -998,6 +1512,33 @@ static int service_load(struct connman_service *service)
 
 	g_free(data);
 
+	switch (service->type) {
+	case CONNMAN_SERVICE_TYPE_UNKNOWN:
+	case CONNMAN_SERVICE_TYPE_ETHERNET:
+		break;
+	case CONNMAN_SERVICE_TYPE_WIFI:
+	case CONNMAN_SERVICE_TYPE_WIMAX:
+	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		service->favorite = g_key_file_get_boolean(keyfile,
+				service->identifier, "Favorite", NULL);
+
+		str = g_key_file_get_string(keyfile,
+				service->identifier, "Failure", NULL);
+		if (str != NULL) {
+			service->state = CONNMAN_SERVICE_STATE_FAILURE;
+			service->error = string2error(str);
+		}
+		break;
+	}
+
+	str = g_key_file_get_string(keyfile,
+				service->identifier, "Modified", NULL);
+	if (str != NULL) {
+		g_time_val_from_iso8601(str, &service->modified);
+		g_free(str);
+	}
+
 	str = g_key_file_get_string(keyfile,
 				service->identifier, "Passphrase", NULL);
 	if (str != NULL) {
@@ -1005,6 +1546,9 @@ static int service_load(struct connman_service *service)
 		service->passphrase = str;
 	}
 
+	__connman_ipconfig_load(service->ipconfig, keyfile,
+					service->identifier, "IPv4.");
+
 	g_key_file_free(keyfile);
 
 	return 0;
@@ -1015,6 +1559,7 @@ static int service_save(struct connman_service *service)
 	GKeyFile *keyfile;
 	gchar *pathname, *data = NULL;
 	gsize length;
+	gchar *str;
 
 	DBG("service %p", service);
 
@@ -1039,9 +1584,50 @@ static int service_save(struct connman_service *service)
 	g_free(data);
 
 update:
+	if (service->name != NULL)
+		g_key_file_set_string(keyfile, service->identifier,
+						"Name", service->name);
+
+	switch (service->type) {
+	case CONNMAN_SERVICE_TYPE_UNKNOWN:
+	case CONNMAN_SERVICE_TYPE_ETHERNET:
+		break;
+	case CONNMAN_SERVICE_TYPE_WIFI:
+	case CONNMAN_SERVICE_TYPE_WIMAX:
+	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		g_key_file_set_boolean(keyfile, service->identifier,
+					"Favorite", service->favorite);
+
+		if (service->state == CONNMAN_SERVICE_STATE_FAILURE) {
+			const char *failure = error2string(service->error);
+			if (failure != NULL)
+				g_key_file_set_string(keyfile,
+							service->identifier,
+							"Failure", failure);
+		} else {
+			g_key_file_remove_key(keyfile, service->identifier,
+							"Failure", NULL);
+		}
+		break;
+	}
+
+	str = g_time_val_to_iso8601(&service->modified);
+	if (str != NULL) {
+		g_key_file_set_string(keyfile, service->identifier,
+							"Modified", str);
+		g_free(str);
+	}
+
 	if (service->passphrase != NULL)
 		g_key_file_set_string(keyfile, service->identifier,
 					"Passphrase", service->passphrase);
+	else
+		g_key_file_remove_key(keyfile, service->identifier,
+							"Passphrase", NULL);
+
+	__connman_ipconfig_save(service->ipconfig, keyfile,
+					service->identifier, "IPv4.");
 
 	data = g_key_file_to_data(keyfile, &length, NULL);
 
